module CURIC::ZoomThrough
  class ZoomThroughOb
    attr_accessor :active_tool, :last_tool_name, :last_camera, :camram, :disable

    def initialize
      @active_tool = ''
      @last_tool_name = ''
      @last_camera = {}
      @camram = {}
      @disable = false

      Sketchup.add_observer(ZoomThroughOb_AppOb.new(self))
    end

    def store_camera(view)
      camera = view.camera
      @last_camera[:eye] = camera.eye
      @last_camera[:target] = camera.target
      @last_camera[:up] = camera.up
      @last_camera[:direction] = camera.direction
    end

    # App
    class ZoomThroughOb_AppOb < Sketchup::AppObserver
      def initialize(observers)
        @instance = observers

        begin
          attach_observers(Sketchup.active_model)
        rescue StandardError
        end
      end

      def onNewModel(model)
        attach_observers(model)
      end

      def onOpenModel(model)
        attach_observers(model)
      end

      def attach_observers(model)
        model.tools.add_observer(ZoomThroughOb_ToolsOb.new(@instance))
        model.active_view.add_observer(ZoomThroughOb_ViewOb.new(@instance))
      end
    end

    # Tools
    class ZoomThroughOb_ToolsOb < Sketchup::ToolsObserver
      def initialize(observers)
        @instance = observers
      end

      def onActiveToolChanged(tools, tool_name, tool_id)
        tool_name = fix_mac_tool_name(tool_name)

        @instance.active_tool = tool_name

        # Lưu vị trí camera thay đổi tool
        @instance.store_camera(tools.model.active_view)

        tool_changed(tool_name, tool_id) unless tool_id.zero?
      end

      def fix_mac_tool_name(tool_name)
        if tool_name == 'eTool'
          tool_name = 'ScaleTool'
        elsif tool_name == 'ool'
          tool_name = 'MoveTool'
        elsif tool_name == 'onentCSTool'
          tool_name = 'ComponentCSTool'
        elsif tool_name == 'PullTool'
          tool_name = 'PushPullTool'
        end
        tool_name
      end

      def tool_changed(tool_name, _tool_id)
        @instance.camram[tool_name] = []
        @instance.last_tool_name = tool_name
      end
    end

    # View
    class ZoomThroughOb_ViewOb < Sketchup::ViewObserver
      def initialize(observers)
        @instance = observers
      end

      def onViewChanged(view)
        return if @instance.disable
        # p "view changed: #{@instance.last_tool_name}"

        camera = view.camera
        direction = camera.direction

        cam = {}
        cam[:eye] = camera.eye
        cam[:up] = camera.up
        cam[:target] = camera.target
        cam[:direction] = camera.direction

        item = view.model.raytest([camera.eye, camera.direction])
        if item && item[1].last.is_a?(Sketchup::Face)
          point, path = item

          cam[:near_face] = {}
          cam[:near_face][:eye_to_face] = camera.eye.distance(point)
          cam[:near_face][:path] = path
        end

        @instance.camram ||= {}
        @instance.camram[@instance.last_tool_name] ||= []
        @instance.camram[@instance.last_tool_name] << cam

        # Kiểm tra có phải đang zoom dựa vào 5 trạng thái cam gần nhất
        if on_zoom? && near_face?(view)
          # Raytest eye to target
          # p 'on_zoom? && near_face'
          cam_data = @instance.camram[@instance.last_tool_name][-1]
          move_on_camera(view, cam_data)
        end

        @instance.store_camera(view)
      end

      def move_on_camera(view, cam_data, n = 1)
        return unless cam_data
        return unless cam_data[:near_face]

        camera = view.camera

        eye = cam_data[:eye]
        target = cam_data[:target]
        up = cam_data[:up]

        direction = cam_data[:direction]
        vector = on_zoom_in?(view) ? direction : direction.reverse
        distance = cam_data[:near_face][:eye_to_face] + 1.mm

        # p "#{self}.move_on_camera(#{n}): #{distance}"

        eye = eye.offset(vector, distance)

        camera.set(eye, target, up)

        item = view.model.raytest([camera.eye, camera.direction])
        return unless item && item[1].last.is_a?(Sketchup::Face)
          
        point, path = item
        d = camera.eye.distance(point)
        return unless d < CURIC::ZoomThrough.read_distance

        cam_data[:eye] = camera.eye
        # cam_data[:near_face] = {}
        cam_data[:near_face][:eye_to_face] = d
        cam_data[:near_face][:path] = path

        move_on_camera(view, cam_data, n + 1)
      end

      def on_zoom?
        datas = @instance.camram[@instance.last_tool_name][-2..-1]
        return unless datas

        ups = datas.map { |c| c[:up] }
        directions = datas.map { |c| c[:direction] }

        zoom = ups.map { |v| v.to_a.map{ |i| i.round(5) } }.uniq.length == 1 && 
               directions.map { |v| v.to_a.map{ |i| i.round(5) } }.uniq.length == 1

        zoom
      end

      def on_zoom_in?(view)
        camera = view.camera
        cam0 = @instance.camram[@instance.last_tool_name][-2]
        cam1 = @instance.camram[@instance.last_tool_name][-1]

        return true unless cam0 && cam1

        eye0 = cam0[:eye]
        eye1 = cam1[:eye]
        vec = eye0.vector_to(eye1)
        return true unless vec.valid?
        
        vec.angle_between(camera.direction).radians < 90
      end

      def on_zoom_out?
        !on_zoom_in?
      end

      def near_face?(view)
        datas = @instance.camram[@instance.last_tool_name][-2..-1]
        return unless datas

        near_faces = datas.map { |c| c[:near_face] }.compact
        return if near_faces.empty?

        # zoom_in thì same faces
        # zoom_out thì ko nhất thiết
        if on_zoom_in?(view)
          same_path = near_faces.all? { |d| d[:path] == near_faces[0][:path] }
          return unless same_path
        else
          # return true
        end

        eye_to_faces = near_faces.map { |d| d[:eye_to_face] }
        # if eye_to_faces[0] > eye_to_faces[-1]
        #   p 'zoom in'
        # else
        #   p 'zoom out'
        # end

        return true if eye_to_faces[-1] < CURIC::ZoomThrough.read_distance

        # kiểm tra face trước mặt có che phần lớn tại con chuột trên màn hình
        eye0 = datas[0][:eye]
        eye1 = datas[1][:eye]
        vec = eye0.vector_to(eye1)
        return unless vec.valid?
          
        mouse = view.screen_coords(eye0.offset(vec, 10))
        # p "mouse: #{mouse.to_a}"

        false
      end
    end
  end
end
